
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "blocks.h"
#include "../Share/global.h"
#include "../Share/csp.h"

/*
 *  Private routines.
 */
static void	set_domain();
static void	set_value_attributes();


/*
 *  Generate a CSP corresponding to a blocks world planning problem.
 */
CSP_type *
generate( time )
    int         time;
{
    CSP_type    *C;
    int         i, a, b, t, t0, t1, t2, v;
    int		next_con;

    C = (CSP_type *)new_calloc( 1, sizeof(CSP_type) );

    C->time = time;
    C->n = (time+1) * (2*get_num_blocks() + 1);

    /*
     *  Allocate space for n variables.
     */
    C->domain_size = new_calloc( C->n+1, sizeof(int) );
    C->variable_attribute = (VARA *)new_calloc( C->n+1, sizeof(VARA) );
    C->value_attribute = (VALA **)new_calloc( C->n+1, sizeof(VALA *) );

    /*
     *  Record the attributes of each variable.
     *  Initialize the domain of each variable.
     */
    for( t = 0; t <= time; t++ ) {
	/*
	 *  Block variables
	 */
	for( i = 1; i <= get_num_blocks(); i++ ) {
	    v = get_var( C, t, ABOVE, i );
	    set_variable_attributes( C, v, ABOVE, t, i );
	    set_domain( C, v );

	    v = get_var( C, t, BELOW, i );
	    set_variable_attributes( C, v, BELOW, t, i );
	    set_domain( C, v );
	}
	/*
	 *  Arm variable
	 */
	v = get_var( C, t, ARM, 1 );
	set_variable_attributes( C, v, ARM, t, 1 );
	set_domain( C, v );
    }

    /*
     *  Is the variable type visible? (1) or hidden? (0)
     */
    C->visible[ABOVE] = 1;
    C->visible[BELOW] = 1;
    C->visible[ARM]   = 0;
    C->n_visible = (time+1) * 2 * get_num_blocks();


    next_con = 0;

    /*
     *  Record the arity, scheme, and relation of a
     *  binary distance constraint on blocks.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
	for( a = 1; a <= get_num_blocks(); a++ ) if( (t2-t1) < 4 ) {
	    set_constraint( C, next_con, 2, DISTANCE_ABOVE_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t1, ABOVE, a );
	    C->scheme[next_con][1] = get_var( C, t2, ABOVE, a );
	    next_con++;
	    set_constraint( C, next_con, 2, DISTANCE_BELOW_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t1, BELOW, a );
	    C->scheme[next_con][1] = get_var( C, t2, BELOW, a );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  4-ary distance constraint on blocks.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time && (t2-t1) < 8; t2++ ) {
	for( a = 1;   a <= get_num_blocks(); a++ )
	for( b = a+1; b <= get_num_blocks(); b++ ) {
	    set_constraint( C, next_con, 4, DISTANCE2_ABOVE_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t1, ABOVE, a );
	    C->scheme[next_con][1] = get_var( C, t2, ABOVE, a );
	    C->scheme[next_con][2] = get_var( C, t1, ABOVE, b );
	    C->scheme[next_con][3] = get_var( C, t2, ABOVE, b );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  4-ary distance constraint on blocks.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time && (t2-t1) < 6; t2++ ) {
	for( a = 1;   a <= get_num_blocks(); a++ )
	for( b = a+1; b <= get_num_blocks(); b++ ) {
	    set_constraint( C, next_con, 4, DISTANCE2_BELOW_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t1, BELOW, a );
	    C->scheme[next_con][1] = get_var( C, t2, BELOW, a );
	    C->scheme[next_con][2] = get_var( C, t1, BELOW, b );
	    C->scheme[next_con][3] = get_var( C, t2, BELOW, b );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  binary constraint on blocks.
     */
    for( t = 1; t <= time-1; t++ ) {
	for( a = 1; a <= get_num_blocks(); a++ ) {
	    set_constraint( C, next_con, 2, SAME_BLOCK_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t, ABOVE, a );
	    C->scheme[next_con][1] = get_var( C, t, BELOW, a );
	    next_con++;
	    if( !clear_in_init( a ) ) {
		b = get_init_above( a );
		set_constraint( C, next_con, 2, DIFF_BLOCK_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t, ABOVE, a );
		C->scheme[next_con][1] = get_var( C, t, BELOW, b );
		next_con++;
	    }
	    if( !clear_in_goal( a ) ) {
		b = get_goal_above( a );
		set_constraint( C, next_con, 2, DIFF_BLOCK_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t, ABOVE, a );
		C->scheme[next_con][1] = get_var( C, t, BELOW, b );
		next_con++;
	    }
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  4-ary constraint on blocks and the arm.
     */
    for( t = 0; t < time; t++ ) {
	for( a = 1; a <= get_num_blocks(); a++ ) {
	    set_constraint( C, next_con, 4, ACTION_ABOVE_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t,   ABOVE, a );
	    C->scheme[next_con][1] = get_var( C, t,   ARM,   1 );
	    C->scheme[next_con][2] = get_var( C, t+1, ABOVE, a );
	    C->scheme[next_con][3] = get_var( C, t+1, ARM,   1 );
	    next_con++;
	    set_constraint( C, next_con, 4, ACTION_BELOW_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t,   BELOW, a );
	    C->scheme[next_con][1] = get_var( C, t,   ARM,   1 );
	    C->scheme[next_con][2] = get_var( C, t+1, BELOW, a );
	    C->scheme[next_con][3] = get_var( C, t+1, ARM,   1 );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of each
     *  2-ary constraint on blocks.
     */
    for( t0 =    0; t0 <= time; t0++ )
    for( t1 = t0+1; t1 <= time; t1++ ) {
	for( a = 1; a <= get_num_blocks(); a++ ) {
	  if( !clear_in_init( a ) &&
	      !clear_in_goal( a ) &&
	      get_init_above( a ) != get_goal_above( a ) ) {
	    set_constraint( C, next_con, 2, USELESS2_ACTIONS_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t0, ABOVE, a );
	    C->scheme[next_con][1] = get_var( C, t1, ABOVE, a );
	    next_con++;
	  }
	  if( !on_table_in_init( a ) &&
	      !on_table_in_goal( a ) &&
	      get_init_below( a ) != get_goal_below( a ) ) {
	    set_constraint( C, next_con, 2, USELESS2_ACTIONS_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t0, BELOW, a );
	    C->scheme[next_con][1] = get_var( C, t1, BELOW, a );
	    next_con++;
	  }
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  3-ary constraint on blocks.
     */
    for( t0 =    0; t0 <= time; t0++ )
    for( t1 = t0+2; t1 <= time; t1++ ) {
	for( a = 1; a <= get_num_blocks(); a++ ) {
	    if( !clear_in_init( a ) &&
	        get_init_above( a ) == get_goal_above( a ) ) {
		set_constraint( C, next_con, 3, USELESS3_ACTIONS_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t0,   ABOVE, a );
		C->scheme[next_con][1] = get_var( C, t0+1, ABOVE, a );
		C->scheme[next_con][2] = get_var( C, t1,   ABOVE, a );
		next_con++;
	    }
	    set_constraint( C, next_con, 3, USELESS4_ACTIONS_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t0,   BELOW, a );
	    C->scheme[next_con][1] = get_var( C, t0+1, BELOW, a );
	    C->scheme[next_con][2] = get_var( C, t1,   BELOW, a );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of each
     *  5-ary constraint on blocks.
     */
    for( t0 =    0; t0 <= time; t0++ ) if( (t0 % 2) == 0 )
    for( t1 = t0+2; t1 <  time; t1++ ) if( (t1 % 2) != 0 ) {
	for( a = 1;   a <= get_num_blocks(); a++ )
	for( b = a+1; b <= get_num_blocks(); b++ ) {
	    set_constraint( C, next_con, 5, SYMMETRY_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t0,   ABOVE, a );
	    C->scheme[next_con][1] = get_var( C, t0,   ABOVE, b );
	    C->scheme[next_con][2] = get_var( C, t0+1, ABOVE, b );
	    C->scheme[next_con][3] = get_var( C, t1,   BELOW, a );
	    C->scheme[next_con][4] = get_var( C, t1+1, BELOW, a );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of
     *  3-ary and 4-ary constraints on blocks.
     */
    for( t = 0; t < time; t++ ) if( (t % 2) != 0 ) {
	for( a = 1; a <= get_num_blocks(); a++ ) {
	    if( !on_table_in_goal( a ) ) {
		b = get_goal_below( a );
		set_constraint( C, next_con, 3, USEFUL_ACTION_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t,   BELOW, a );
		C->scheme[next_con][1] = get_var( C, t+1, BELOW, a );
		C->scheme[next_con][2] = get_var( C, t+1, BELOW, b );
		next_con++;
	    }
	}
    }

    C->m = next_con;

    /*
     *  When should the constraint be propagated?
     *
     *  C->checkable[r] = u means that constraint type r is propagated
     *          only when the number of uninstantiated variables in the
     *          scheme of the constraint is less than or equal to u.
     *          Constraint propagation then potentially reduces the
     *          domains of each of the uninstantiated variables.
     *  
     *  In particular,
     *    == 1    means only forward checking done on this constraint
     *    >= 2    means arc consistency checking done on this constraint
     *    == HUGE means constraint is always propagated
     */
    C->checkable[DISTANCE_ABOVE_CONSTRAINT] = HUGE;
    C->checkable[DISTANCE_BELOW_CONSTRAINT] = HUGE;
    C->checkable[DISTANCE2_ABOVE_CONSTRAINT] = 1;
    C->checkable[DISTANCE2_BELOW_CONSTRAINT] = 1;
    C->checkable[SAME_BLOCK_CONSTRAINT] = HUGE;
    C->checkable[DIFF_BLOCK_CONSTRAINT] = HUGE;
    C->checkable[ACTION_ABOVE_CONSTRAINT] = HUGE;
    C->checkable[ACTION_BELOW_CONSTRAINT] = HUGE;
    C->checkable[USELESS2_ACTIONS_CONSTRAINT] = 1;
    C->checkable[USELESS3_ACTIONS_CONSTRAINT] = 1;
    C->checkable[USELESS4_ACTIONS_CONSTRAINT] = 1;
    C->checkable[SYMMETRY_CONSTRAINT] = 1;
    C->checkable[USEFUL_ACTION_CONSTRAINT] = HUGE;

    construct_index( C );

    return( C );
}

int
get_var( C, state, type, number )
    CSP_type    *C;
    int		state, type, number;
{
    int		i;

    i = state * (2 * get_num_blocks() + 1);

    if( type == ABOVE ) {
	i += number;
    }
    else
    if( type == BELOW ) {
	i += get_num_blocks() + number;
    }
    else
    if( type == ARM ) {
	i += 2 * get_num_blocks() + number;
    }
    return( i );
}

static void
set_value_attributes( C, v, a, type, number )
    CSP_type    *C;
    int		type, number;
{
    C->value_attribute[v][a].type = type;
    C->value_attribute[v][a].number = number;
}

/*
 *  Initialize the domain of a variable.
 *  Record the attributes of each value.
 */
static void
set_domain( C, v )
    CSP_type    *C;
    int		v;
{
    int		state, type, i, j;
    int		next_val;

    get_variable_attributes( C, v, &type, &state, &i );
    if( type == ABOVE ) {
	if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    if( clear_in_init( i ) )
		set_value_attributes( C, v, 0, CLEAR, 0 );
	    else
		set_value_attributes( C, v, 0, BLOCK, get_init_above( i ) );
	    C->domain_size[v] = 1;
	}
	else
	if( state == C->time ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    if( clear_in_goal( i ) )
		set_value_attributes( C, v, 0, CLEAR, 0 );
	    else
		set_value_attributes( C, v, 0, BLOCK, get_goal_above( i ) );
	    C->domain_size[v] = 1;
	}
	else {
	    /*
	     * Domains for ABOVE variables: 0, 1 or 2 blocks, CLEAR, HELD.
	     */
            C->value_attribute[v] = (VALA *)new_calloc( 4, sizeof(VALA) );
	    next_val = 0;
	    if( !clear_in_init( i ) &&
		get_init_above( i ) == get_goal_above( i ) ) {
	        set_value_attributes( C, v, next_val, BLOCK, get_init_above( i ) );
		next_val++;
	    }
	    else {
		if( !clear_in_init( i ) &&
		    (C->time-state) >= 2*(get_goal_depth( i )-1)+2 ) {
		    set_value_attributes( C, v, next_val, BLOCK, get_init_above( i ) );
		    next_val++;
		}
		if( !clear_in_goal( i ) &&
		    state >= 2*(get_init_depth( i )-1)+2 ) {
		    set_value_attributes( C, v, next_val, BLOCK, get_goal_above( i ) );
		    next_val++;
		}
	    }
	    if(          state  >= 2*(get_init_depth( i )-1)-1 &&
		(C->time-state) >= 2*(get_goal_depth( i )-1)-1 ) {
		set_value_attributes( C, v, next_val, CLEAR, 0 );
		next_val++;
	    }
	    if( (state % 2) != 0 &&
		         state  >= 2*(get_init_depth( i )-1)+1 &&
		(C->time-state) >= 2*(get_goal_depth( i )-1)+1 ) {
		set_value_attributes( C, v, next_val, HELD, 0 );
		next_val++;
	    }
	    C->domain_size[v] = next_val;
	}
    }
    else
    if( type == BELOW ) {
	if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    if( on_table_in_init( i ) )
		set_value_attributes( C, v, 0, TABLE, 0 );
	    else
		set_value_attributes( C, v, 0, BLOCK, get_init_below( i ) );
	    C->domain_size[v] = 1;
	}
	else
	if( state == C->time ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    if( on_table_in_goal( i ) )
		set_value_attributes( C, v, 0, TABLE, 0 );
	    else
		set_value_attributes( C, v, 0, BLOCK, get_goal_below( i ) );
	    C->domain_size[v] = 1;
	}
	else {
	    /*
	     * Domains for BELOW variables: 0, 1 or 2 blocks, TABLE, HELD.
	     */
            C->value_attribute[v] = (VALA *)new_calloc( 4, sizeof(VALA) );
	    next_val = 0;
	    if( !on_table_in_init( i ) &&
		get_init_below( i ) == get_goal_below( i ) ) {
	        set_value_attributes( C, v, next_val, BLOCK, get_init_below( i ) );
		next_val++;
	    }
	    else {
		if( !on_table_in_init( i ) &&
		    (C->time-state) >= 2*(get_goal_depth( i )-1)+2 ) {
		    set_value_attributes( C, v, next_val, BLOCK, get_init_below( i ) );
		    next_val++;
		}
		if( !on_table_in_goal( i ) &&
		    state >= 2*(get_init_depth( i )-1)+2 ) {
		    set_value_attributes( C, v, next_val, BLOCK, get_goal_below( i ) );
		    next_val++;
		}
	    }
	    if( (on_table_in_init( i ) ||          state  >= 2*get_init_depth( i )) &&
		(on_table_in_goal( i ) || (C->time-state) >= 2*get_goal_depth( i )) ) {
		set_value_attributes( C, v, next_val, TABLE, 0 );
		next_val++;
	    }
	    if( (state % 2) != 0 &&
		         state  >= 2*(get_init_depth( i )-1)+1 &&
		(C->time-state) >= 2*(get_goal_depth( i )-1)+1 &&
		(!on_table_in_init( i ) || !on_table_in_goal( i )) ) {
		set_value_attributes( C, v, next_val, HELD, 0 );
		set_value_attributes( C, v, next_val, HELD, 0 );
		next_val++;
	    }
	    C->domain_size[v] = next_val;
	}
    }
    else
    if( type == ARM ) {
	/*
	 * Arm variable.
	 * Domain is all blocks or EMPTY.
	 */
	if( (state % 2) == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    set_value_attributes( C, v, 0, EMPTY, 0 );
	    C->domain_size[v] = 1;
	}
	else {
            C->value_attribute[v] = (VALA *)new_calloc( get_num_blocks(), sizeof(VALA) );
	    for( j = 1; j <= get_num_blocks(); j++ ) {
	        set_value_attributes( C, v, j-1, BLOCK, j );
	    }
	    C->domain_size[v] = get_num_blocks();
	}
    }
}

